/*
  * Copyright (c) 2019 Ignacio Francisco Ramírez Paulino and Ignacio Hounie
  * This program is free software: you can redistribute it and/or modify it
  * under the terms of the GNU Affero General Public License as published by
  * the Free Software Foundation, either version 3 of the License, or (at
  * your option) any later version.
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero
  * General Public License for more details.
  * You should have received a copy of the GNU Affero General Public License
  *  along with this program. If not, see <http://www.gnu.org/licenses/>.
*/
/**
 * \file paco_dct_inpaint_tool.c
 * \brief Command line interface to the PACO-DCT inpainting algorithm.
 *
 *  Uses GNU argp to parse arguments
 *  see http://www.gnu.org/software/libc/manual/html_node/Argp.html
 */
#include <stdlib.h>
#include <argp.h>                     // argument parsing
#include <string.h>
#include "paco_log.h"                 // logging utilities
#include "paco_io.h"                  // image I/O
#include "paco_metrics.h"                  //performance metrics
/**
 * These are the options that we can handle through the command line
 */
static struct argp_option options[] = {
   { 0 } // terminator
};

/**
 * PACO-DCT and ADMM options. These are filled in by the argument parser
 */
typedef struct  {
    char *input_file;  ///< input (degraded) image
    char *mask_file;   ///< inpainting mask: nonzero values are considered erased
    char *ref_file;    ///< if available, this will be used to assess the quality of the result
} paco_config_st; 

/**
 * options handler
 */
static error_t parse_opt (int key, char *arg, struct argp_state *state);

/**
 * General description of what this program does; appears when calling with --help
 */
static char program_doc[] =
  "\n*** paco_compare_tool: compare two images, the first is assumed to be         ***\
   \n*** a reference or ground truth image, and the second is a distorted version. ***\
   \n*** A third, optional mask indicates which pixels are to be compared.         ***\
   \n*** Results are given in terms of the  Root Mean Squared Error (RMSE)         ***\
   \n*** and Structural Similarity Metric (SSIM).                                  ***";

/**
 * A general description of the input arguments we accept; appears when calling with --help
 */
static char args_doc[] = "<INPUT> <REF> <MASK>";

/**
 * argp configuration structure
 */
static struct argp argp = { options, parse_opt, args_doc, program_doc };

/**
 * main function
 */
int main(int argc, char **argv) {
    paco_config_st cfg; // command-line program configuration


    /*
     * Default program configuration
     */
    set_log_level(PACO_LOG_INFO);
    set_log_stream(stdout);
    /*
     * call parser
     */
    argp_parse (&argp, argc, argv, 0, 0, &cfg);
   /*
     * load data.
     */
    paco_image_st* input = NULL;
    paco_image_st* ref  = NULL;
    paco_image_st* mask  = NULL;

    input = read_png_file(cfg.input_file);
    if(!input) {
        paco_error("Unable to read input!\n");
        exit(1);
    }
    /*
     * load ref file (mandatory) 
     */
    ref = read_png_file(cfg.ref_file);
    if(!ref) {
        paco_error("Unable to read ref!\n");
        exit(1);
    }
     /*
     * load mask file (mandatory) 
     */
    mask = read_png_file(cfg.mask_file);
    if(!mask) {
        paco_error("Unable to read mask!\n");
        exit(1);
    }

   const index_t nchannels = get_nchannels(input);
    for (index_t c = 0; c < nchannels; c++) {
        /*
         * channels are identified by a single letter
         */
        const char channel_name = get_channel_name(input,c);
        /*
         * get c-th channel from image
         */
        gsl_matrix_float* input_samples = get_channel_samples(input,c);
        gsl_matrix_float* ref_samples = get_channel_samples(ref,c);
        gsl_matrix_float* mask_samples = get_channel_samples(mask,0);
        float* ref_data = ref_samples->data;
        float* mask_data = mask_samples->data;
        float* input_data = input_samples->data;
        const index_t tda = input_samples->tda;
        const index_t ncols = input_samples->size2;
        const index_t nrows = input_samples->size1;
        double rmse = rmse_partial(ref_data, input_data, mask_data,nrows*ncols);
        double ssim =  iqa_ms_ssim(ref_data, input_data, mask_data, ncols, nrows, tda, NULL);
        paco_info("CHANNEL %c RMSE: %f      SSIM: %f \n", channel_name, 255*rmse,  ssim);
    }
    if (nchannels > 1) {
        paco_image_convert_colorspace(input,COLORSPACE_YUV);
        paco_image_convert_colorspace(ref,COLORSPACE_YUV);
        for (index_t c = 0; c < nchannels; c++) {
            /*
            * channels are identified by a single letter
            */
            const char channel_name = get_channel_name(input,c);
            /*
            * get c-th channel from image
            */
            gsl_matrix_float* input_samples = get_channel_samples(input,c);
            gsl_matrix_float* ref_samples = get_channel_samples(ref,c);
            gsl_matrix_float* mask_samples = get_channel_samples(mask,0);
            float* ref_data = ref_samples->data;
            float* mask_data = mask_samples->data;
            float* input_data = input_samples->data;
            const index_t tda = input_samples->tda;
            const index_t ncols = input_samples->size2;
            const index_t nrows = input_samples->size1;
            double rmse = rmse_partial(ref_data, input_data, mask_data,nrows*ncols);
            double ssim =  iqa_ms_ssim(ref_data, input_data, mask_data, ncols, nrows, tda, NULL);
            paco_info("CHANNEL %c RMSE: %f      SSIM: %f \n", channel_name, 255*rmse,  ssim);
        }
    }
   /*
     * free local storage.
     */
    paco_image_free(input);
    free(input);
    paco_image_free(mask);
    free(mask);
    paco_image_free(ref);
    free(ref);
    exit (0);
}


/*
 * argp callback for parsing a single option.
 */
static error_t parse_opt (int key, char *arg, struct argp_state *state) {
    /* Get the input argument from argp_parse,
     * which we know is a pointer to our arguments structure.
     */
    paco_config_st *cfg = (paco_config_st*) state->input;
    switch (key) {
    case ARGP_KEY_ARG:
        switch (state->arg_num) {
        case 0:
            cfg->input_file = arg;
            break;
        case 1:
            cfg->ref_file = arg;
            break;
        case 2:
            cfg->mask_file = arg;
            break;
        default:
            /** too many arguments! */
            paco_error("Too many arguments!.\n");
            argp_usage (state);
            break;
        }
        break;
    case ARGP_KEY_END:
        if (state->arg_num < 3) {
            /* Not enough mandatory arguments! */
            paco_error("Too FEW arguments!\n");
            argp_usage (state);
        }
        break;

    default:
        return ARGP_ERR_UNKNOWN;
    }
    return 0;
}
