/*
  * Copyright (c) 2019 Ignacio Francisco Ramírez Paulino and Ignacio Hounie
  * the Free Software Foundation, either version 3 of the License, or (at
  * your option) any later version.
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero
  * General Public License for more details.
  * You should have received a copy of the GNU Affero General Public License
  *  along with this program. If not, see <http://www.gnu.org/licenses/>.
*/
/**
 * \file paco_dct_inpainting_problem.c
 * \brief Implementation of the PACO-DCT inpainting problem  
 *
 */
#include "paco_dct_inpainting_problem.h"
#include <math.h>
#include <assert.h>
#include <string.h>
#include "paco_log.h"
#include "paco_linop.h"
#include "paco_io.h"
#include "paco_metrics.h"
#include "paco_linop.h"
#include <string.h>
#include "paco_image.h"

static gsl_vector_float* weights; ///< DCT coefficient weights
static const gsl_matrix_float* input_samples; ///< input signal samples (one channel)
static gsl_matrix_float* stitched_samples; ///< auxiliary signal; contains the result of the algorithm at the end
static const gsl_matrix_float* mask_samples; ///< inpainting mask samples
static const gsl_matrix_float* ref_samples; ///< reference signal samples, if available

static index_t* idx_missing_samples; ///< missing pixels' linear indexes
static index_t num_missing_samples; ///< number of missing pixels; computed.
static index_t* idx_projected_samples;  ///< projected pixels linear indexes: this is the set O in the paper
static index_t num_projected_samples; ///< affected pixels with known values: this is the size of the set O in the paper

static char iter_prefix[1024]; ///< for saving intermediate results

//============================================================================

/**
 * weighted l1 cost function: f(X) = \sum_{i=1}^{n}\sum_{j=1}^{m} w_j |X_{ij} |
 */
double paco_eval_f(const gsl_matrix_float* X) {
    const gsl_vector_float* W = weights;
    assert(W->size == X->size2);
    const index_t m = X->size1;
    const index_t n = X->size2;
    const size_t    sX = X->tda;

    double val = 0;
#ifdef _OPENMP
    #pragma omp parallel for reduction(+:val)
#endif
    for (index_t i = 0; i < m; ++i) {
        const sample_t* pW = W->data;
        const sample_t* pXi = X->data + i*sX;
        for (index_t j = 0; j < n; ++j, pW++, pXi++) {
            val += *pW *fabs(*pXi);
        }
    }
    return val/X->size1/X->size2;
}

//============================================================================

/**
 * \brief Soft thresholding operator defined in EQUATION (11) in the paper.
 *
 * Given a matrix A and weights W
 * returns a matrix SA with the same dimensions as A in which each element is given by
 * SA_{i,j} = A_{i,j} - tau*W_{j} If A_{i,j} > tau
 *          = A_{i,j} + tau*W_{j} If A_{i,j} < -tau
 *          = SA_{i,j} = 0        Otherwise (-tau < A_{i,j} < tau)
 * @param  A   n x m input matrix,  W  1 x m weights, tau
 * @return SA  n x m matrix (preallocated)
 */

void paco_prox_f( gsl_matrix_float *SA, const gsl_matrix_float* A, const double tau) {
    paco_debug("soft thresholding.\n");
    assert(SA->size1 == A->size1);
    assert(SA->size2 == A->size2);
    const index_t m = A->size1;
    const index_t n = A->size2;

    const gsl_vector_float* W = weights;

    const sample_t* pA = A->data;
    const size_t    sA = A->tda;
    sample_t* pSA = SA->data;
    const size_t    sSA = SA->tda;
    const sample_t* pW  = W->data;

#ifdef _OPENMP
    #pragma omp parallel for shared(pW)
#endif
    for (int i = 0; i < m; ++i) {
        const sample_t* pAT = pA + i*sA;
        sample_t* pSAT = pSA + i*sSA;
        for (int j= 0; j < n; ++j) {
            const sample_t twj = tau*pW[j]; //gsl_vector_float_get(W,j);
            const sample_t Aij = pAT[j]; //gsl_matrix_float_get(A,i,j);
            if (Aij > twj) {
                pSAT[j] = Aij - twj; //gsl_matrix_float_set(SA, i, j, Aij - twj);
            } else if (Aij < -twj) {
                pSAT[j] = Aij + twj; //gsl_matrix_float_set(SA, i, j, Aij + twj);
            } else {
                pSAT[j] = 0; //gsl_matrix_float_set(SA, i, j, 0);
            }
        }
    }
}

//============================================================================

/**
 * \brief Proximal operator for the PACO-DCT inpainting problem (ALGORITHM 3 in the paper.)
 * It performs the following steps:
 * 1) convert patches from DCT to sample space using the IDCT
 * 2) stitch the patches, creating an image I
 * 3) Fills in known sample values from the image I, creating PI
 * 4) Extracts the patches from PI
 * 5) Convert back the patches to DCT space by applying the direct DCT to them.DCT_H
 *
 * @param[out] proxA output matrix with the result of the proximal operator
 * @param A input to the proximal operator
 * @param lambda parameter of the proximal operator (not used in this case)
 * @return the result of the above procedure
 */
void paco_prox_g(gsl_matrix_float* PB, const gsl_matrix_float* B, const double lambda) {
    paco_debug("prox_dct_inpainting.\n");
    const  gsl_matrix_float* X = input_samples;
    gsl_matrix_float* V = stitched_samples;
    if (B != PB) {
        gsl_matrix_float_memcpy(PB, B);
    }
    const index_t m = B->size2;
    const index_t w = (index_t) sqrt(m); // only square patches for now
    idct2dn(PB,w,w,PB);
//* 2) stitch the patches, creating an image V
    stitch_patches(V, PB);
//* 3) Overwrites samples in V with known samples from O,
    sample_t* pV = V->data;
    const sample_t* pX = X->data;
    for (int k = 0; k < num_projected_samples; ++k) {
        pV[idx_projected_samples[k]] = pX[idx_projected_samples[k]];
    }
//* 4) Extracts the patches from V
    extract_patches(PB,V);
//* 5) Convert back the patches to DCT space by applying the direct DCT to them.DCT_H
    dct2dn(PB,w,w,PB);
}

//============================================================================

/**
 * as this is an indicator function, and B is always feasible, the value is always 0 
 */
double paco_eval_g(const gsl_matrix_float* B) {
    return 0;
}

//============================================================================

/**
 * writes the intermediate images corresponding to the main (A) and split (B)
 * iterates of the algorithm. 
 */
void paco_sentinel(paco_iterate_st* iter) {
    if ( (iter->k > 0) && (1 << (int)log2(iter->k)) != iter->k ) {
        return;
    }
    const index_t pw = get_patch_width();
    const index_t ph = get_patch_height();
    char aux[128];
    paco_image_st* iter_img;

    //
    // pseudo-image for multiplier U
    //

    idct2dn(iter->U,ph,pw,iter->U);
    stitch_patches(iter->Y,iter->U);
    dct2dn(iter->U,ph,pw,iter->U);
    
    if (strlen(iter_prefix)) {
        snprintf(aux,128,"%s%05d_U.png",iter_prefix,iter->k);
        iter_img = paco_image_from_samples(iter->Y,COLORSPACE_GRAY);
        paco_image_norm(iter_img);
        write_png_file(aux,iter_img);
        paco_image_free(iter_img);
        free(iter_img);
    }

    //
    // partial result for main variable A
    //

    idct2dn(iter->A,ph,pw,iter->A);
    stitch_patches(iter->Y,iter->A);
    dct2dn(iter->A,ph,pw,iter->A);

    if (strlen(iter_prefix)) {
        snprintf(aux,128,"%s%05d_A.png",iter_prefix,iter->k);
        iter_img = paco_image_from_samples(iter->Y,COLORSPACE_GRAY);
        paco_image_norm(iter_img);
        write_png_file(aux,iter_img);
        paco_image_free(iter_img);
        free(iter_img);
    }

    //
    // partial result for split variable B
    //

    idct2dn(iter->B,ph,pw,iter->B);
    stitch_patches(iter->Y,iter->B);
    dct2dn(iter->B,ph,pw,iter->B);

    if (strlen(iter_prefix)) {
        snprintf(aux,128,"%s%05d_B.png",iter_prefix,iter->k);
        iter_img = paco_image_from_samples(iter->Y,COLORSPACE_GRAY);
        paco_image_norm(iter_img);
        write_png_file(aux,iter_img);
        paco_image_free(iter_img);
        free(iter_img);
    }

    //
    // the partial result from B is also the  
    // partial output of the algorithm
    // thus, we compute partial stats from it.
    //
    if (ref_samples) {
        float* out_data = iter->Y->data;
        float* mask_data = mask_samples->data;
        float* ref_data = ref_samples->data;
        const index_t tda = input_samples->tda;
        const index_t ncols = input_samples->size2;
        const index_t nrows = input_samples->size1;
        double rmse = rmse_partial(ref_data, out_data, mask_data, nrows*ncols);
        double ssim =  iqa_ms_ssim(ref_data, out_data, mask_data, ncols, nrows, tda, NULL);
        paco_info("RMSE: %f      SSIM: %f \n", 255*rmse,  ssim);
    }

}

//============================================================================

/**
 * provide an initial value for the PACO-DCT ADMM algorithm 
 */
void paco_iter_init(paco_iterate_st* iter, const gsl_matrix_float* initial_samples) {
    paco_info("Initializing inpainting problem...\n");

    const index_t M = get_input_height();
    const index_t N = get_input_width();
    const index_t w = get_patch_width(); // patch width

    if (initial_samples == NULL) {
        //
        // compute the average value of the known samples
        //
        paco_info("computing average pixel value of known regions.\n");
        double numpix = 0;
        double sumpix = 0;
        for (index_t i = 0; i < M; i++) {
            for (index_t j = 0; j < N; j++) {
                if (gsl_matrix_float_get(mask_samples,i,j) == 0) {
                    sumpix += gsl_matrix_float_get(input_samples,i,j);
                    numpix ++;
                }
            }
        }
        //
        // copy known regions from input
        //
        gsl_matrix_float_memcpy(iter->Y,input_samples);
        //
        // fill unknown regions with average value
        //
        sumpix /= numpix;
        paco_info("initializing erased regions to value %3.0f\n",sumpix);
        for (index_t i = 0; i < M; i++) {
            for (index_t j = 0; j < N; j++) {
                if (gsl_matrix_float_get(mask_samples,i,j) != 0) {
                    gsl_matrix_float_set(iter->Y,i,j,sumpix);
                }
            }
        }
    } else {
        gsl_matrix_float_memcpy(iter->Y,initial_samples);
    }
    //
    // extract the patches from the initial signal estimation
    // and compute their DCT: this is the initial value of the ADMM variable B
    //
    extract_patches(iter->B,iter->Y);
    dct2dn(iter->B,w,w,iter->B);
    //
    // the variable A is initialized to 0
    //
    gsl_matrix_float_set_zero(iter->A);
    //
    // the lagrange multiplier is initialized to 0
    //
    gsl_matrix_float_set_zero(iter->U);
}

//============================================================================

/**
 * set a new reference image
 */
void paco_problem_set_ref(const gsl_matrix_float* _ref) {
    ref_samples = _ref;
}

//============================================================================

/**
 * set a new input image
 */
void paco_problem_set_input(const gsl_matrix_float* _input) {
    input_samples = _input;
}

//============================================================================

/**
 * change file prefix for storing intermediate results
 */
void paco_problem_set_iter_prefix(const char* t) {
    if (t == NULL) {
        iter_prefix[0] = 0;
    } else {
        strncpy(iter_prefix,t,1024);
    }
}

//============================================================================

/**
 * estimate DCT coefficient weights for weighted l1 model
 */
static void paco_problem_estimate_weights() {
    paco_info("estimating dct coefficient weights...\n");
    if (!input_samples) {
        paco_error("Input not defined!");
        return;            
    }
    //
    //count
    //
    const index_t M = get_input_height();
    const index_t N = get_input_width();
    const index_t w1 = get_patch_height();
    const index_t w2 = get_patch_width();
    index_t num_complete = 0;
    for (int i = 0; i <= M - w1; i ++) {
        for (int j= 0; j <= N - w2; j ++) {
            // check that patch is complete
            if (is_patch_complete(mask_samples,i,j,i+w1,j+w2)) {
                num_complete++;
            }
        }
    }
    paco_info("Number of complete patches: %d\n", num_complete);
    //
    // Alloc and extract patches
    //
    gsl_matrix_float* patches = gsl_matrix_float_alloc(num_complete,get_num_patch_samples());
    const index_t patch_width = get_patch_width();
    int patch_idx = 0;
    for (int i = 0; i <= M - w1; i ++) {
        for (int j= 0; j <= N - w2; j ++) {
            // extract patch and check that it is complete
            if (is_patch_complete(mask_samples,i,j,i+w1,j+w2)) {
                //Extract patch
                sample_t* patch_data = patches->data + patches->tda*patch_idx;
                extract_patch(patch_data,input_samples, i, j, patch_width, patch_width);
                patch_idx++;
            }
        }
    }
    assert(patch_idx == num_complete);
    //
    //Compute dct
    //
    dct2dn(patches,w1,w2,patches);

    paco_info("Summing absolute values\n");
    const index_t m = get_num_patch_samples();
    gsl_vector_float_set_zero(weights);
    sample_t* pweights = weights->data;
    double* dctaccu = (double*) calloc( m, sizeof(double) );
    for (int k = 0; k < num_complete; k ++) {
        for (int l = 0; l < m; l ++) {
            pweights[l] += fabs(gsl_matrix_float_get(patches, k, l));
        }
    }
    gsl_matrix_float_free(patches);
    const double minw = gsl_vector_float_min(weights);
    FILE* fw = fopen("weights.txt","w");
    for (int l = 0; l < m; l ++) {
        double wl = pweights[l] > 0 ? minw/pweights[l] : 1.0;
        //double wl = num_complete / (pweights[l] +num_complete * __DBL_EPSILON__);
        pweights[l] = wl;
        fprintf(fw,"%f\n",wl);
    }
    fclose(fw);
    free(dctaccu);
}

//============================================================================

/**
 * set DCT coefficient weights for weightef L1 cost function
 */
void paco_problem_set_weights(const gsl_vector_float* _weights) {
    if (!weights) {
        weights = gsl_vector_float_alloc(get_num_patch_samples());
    }
    if (_weights) {
        const index_t m = _weights->size;
        if (m != get_num_patch_samples()) {
            paco_error("Weighs vector size %d is not compatible with number of DCT coefficients %d",
                (int)m, (int)get_num_patch_samples());
        }
        gsl_vector_float_memcpy(weights,_weights);
    } else {
        paco_problem_estimate_weights();
    }
}

//============================================================================

/**
 * initialize problem structures and state 
 */
void paco_dct_inpainting_problem_init(const gsl_matrix_float* mask) {
    paco_info("Creating inpainting problem...\n");

    mask_samples = mask;
    const index_t M = mask->size1;
    const index_t N = mask->size2;
    const index_t tda = mask->tda;
    stitched_samples = gsl_matrix_float_alloc(M,N);

    paco_debug("input: %d x %d\n",M,N);
    /// Counting missing pixels:
    ///Non zero values are interpreted as missing pixels
    //
    // gather missing pixels information
    //
    int num_mis = 0;
    for (int i = 0; i < M; i ++) {
        for (int j= 0; j < N; j ++) {
            if (gsl_matrix_float_get(mask,i, j) != 0)
                num_mis++;
        }
    }
    num_missing_samples=num_mis;
    paco_debug("Number of missing pixels :%ld\n",num_mis);
    if (num_mis == 0) {
        paco_warn("No missing pixels. Nothing else to do here.");
        idx_missing_samples = 0;
        num_projected_samples = 0;
        idx_projected_samples = 0;
        return;
    }
    idx_missing_samples = (index_t*)malloc(num_mis*sizeof(index_t));
    num_mis = 0;
    for (int i = 0; i < M; i ++) {
        for (int j= 0; j < N; j ++) {
            if (gsl_matrix_float_get(mask, i, j)) {
                idx_missing_samples[num_mis++] = i*tda + j;
            }
        }
    }
    //
    // gather projected pixels information
    // these are the affected pixels which are not missing
    //
    {
        const index_t  nmapped = get_num_mapped_samples();
        const index_t* pmapped = get_idx_mapped_samples();
        const index_t* pmis = idx_missing_samples;
        paco_info("Number of missing pixels: %ld  \n", num_mis);
        paco_info("Number of mapped pixels: %ld\n", nmapped);
        index_t imis = 0;
        index_t nproj = 0;
        index_t next_missing = pmis[imis];
        //Count projected pixels
        for (index_t imap = 0; imap < nmapped; imap++) {
            const index_t idx_map = pmapped[imap];
            if (idx_map != next_missing) {
                nproj++;
            } else { // hit a missing pixel
                imis++;
                if (imis >= num_mis) {
                    // zero can never be the 'next' missing pixel
                    // so this here means that all the remaining mapped
                    // pixels will be projected
                    next_missing = 0;
                } else {
                    next_missing = pmis[imis];
                }
            }
        }
        //Allocate storage for indexes
        num_projected_samples = nproj;
        idx_projected_samples = (index_t*) malloc(nproj*sizeof(index_t));
        paco_info("Number of projected pixels: %ld\n", nproj);
        if (imis < num_mis) paco_info("Warning: %d missing pixels were not mapped",  num_mis - imis);

        //Store projected indexes
        index_t iproj = 0;
        imis = 0;
        next_missing = pmis[imis];
        for (index_t imap = 0; imap < nmapped; imap++) {
            const index_t idx_map = pmapped[imap];
            if (idx_map != next_missing) {
                idx_projected_samples[iproj++] = idx_map;
            } else { // hit a missing pixel
                imis++;
                if (imis >= num_mis) {
                    // zero can never be the 'next' missing pixel
                    // so this here means that all the remaining mapped
                    // pixels will be projected
                    next_missing = 0;
                } else {
                    next_missing = pmis[imis];
                }
            }
        }
    }
}


//============================================================================

/**
 * destroy problem structures and free any allocated space 
 */
void paco_problem_destroy() {
    paco_info("Destroying problem...\n");
    free(idx_missing_samples);
    free(idx_projected_samples);
    gsl_matrix_float_free(stitched_samples);
    if (weights) gsl_vector_float_free(weights);
}
