/*
  * Copyright (c) 2019 Ignacio Francisco Ramírez Paulino and Ignacio Hounie
  * This program is free software: you can redistribute it and/or modify it
  * under the terms of the GNU Affero General Public License as published by
  * the Free Software Foundation, either version 3 of the License, or (at
  * your option) any later version.
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero
  * General Public License for more details.
  * You should have received a copy of the GNU Affero General Public License
  *  along with this program. If not, see <http://www.gnu.org/licenses/>.
*/
/**
 * \file paco_dct_inpaint_tool.c
 * \brief Command line interface to the PACO-DCT inpainting algorithm.
 *
 *  Uses GNU argp to parse arguments
 *  see http://www.gnu.org/software/libc/manual/html_node/Argp.html
 */
#include <stdlib.h>
#include <time.h>                     // basic benchmarking
#include <argp.h>                     // argument parsing
#include <string.h>
#include "paco_log.h"                 // logging utilities
#include "paco_io.h"                  // image I/O
#include "paco_admm.h"                // ADMM algoritm
#include "paco_metrics.h"                  //performance metrics
#include "paco_dct_inpainting_problem.h"  // inpainting problem functions and data
#include "paco_grid_mapping.h"        // patch extraction and stitching using a regular grid of patches
/**
 * These are the options that we can handle through the command line
 */
static struct argp_option options[] = {
    {"verbose",        'v', 0,                OPTION_ARG_OPTIONAL,  "Produce verbose output" },
    {"quiet",          'q', 0,                OPTION_ARG_OPTIONAL,  "Don't produce any output" },
    {"width",          'w', "patch-width",    0, "Patch width in both directions" },
    {"stride",         's', "patch-stride",   0, "Patch stride in both directions." },
    {"lambda",         'l', "lambda",         0, "ADMM penalty parameter." },
    {"max-iter",       'm', "max-iter",       0, "Maximum allowed iterations." },
    {"epsilon",        'e', "epsilon",        0, "Minimum change between iterations." },
    {"kappa",          'k', "kappa",          0, "ADMM penalty parameter multiplier (0 < kappa <= 1)." },
    {"dct-weights",    'c', "dct-weights",    0, "Path to file containing DCT weights" },
    {"initialisation", 'i', "initialisation", 0, "Path to file containing initialisation" },
    {"prefix",         'p', "iter-prefix",    0, "Intermediate results prefix (empty to disable saving int. results)." },
    { 0 } // terminator
};

/**
 * PACO-DCT and ADMM options. These are filled in by the argument parser
 */
typedef struct  {
    char *input_file;  ///< input (degraded) image
    char *mask_file;   ///< inpainting mask: nonzero values are considered erased
    char *output_file; ///< output of the algorithm
    char *ref_file;    ///< if available, this will be used to assess the quality of the result
    char *work_dir;    ///< directory where intermediate results will be written
    char *weight_file; ///< if defined and readable, DCT weights will be loaded from this ASCII file
    char *init_file;   ///< if defined and readable, the initial output estimate will be loaded from this file
    char *iter_prefix; ///< if not null or empty, prefix for saving intermediate results (otherwise no intermediate results are saved)
    index_t width;     ///< patch width (both vertical and horizontal)
    index_t stride;    ///< distance between adjacent patches
    double lambda;     ///< initial value of the ADMM parameter
    index_t maxiter;   ///< maximum number of ADMM iterations allowed
    double minchange;  ///< stop ADMM if the change in cost function f() drops below this value between iterations
    double kappa;      ///< if the cost function f() increases instead of decreasing, lambda is shrank by this amount
} paco_config_st; 

/**
 * options handler
 */
static error_t parse_opt (int key, char *arg, struct argp_state *state);

/**
 * General description of what this program does; appears when calling with --help
 */
static char program_doc[] =
    "\n*** PACO-DCT: A simple, fast and effective inpainting program using    ***\
   \n*** Patch Consensus and a weighted l1 prior on patch DCT coefficients. ***";

/**
 * A general description of the input arguments we accept; appears when calling with --help
 */
static char args_doc[] = "<INPUT> <MASK> <OUTPUT> <REF(optional)>  [REFERENCE]";

/**
 * argp configuration structure
 */
static struct argp argp = { options, parse_opt, args_doc, program_doc };

/**
 * main function
 */
int main(int argc, char **argv) {
    paco_config_st cfg; // command-line program configuration
    time_t t0, t1;      // for benchmarking

    paco_info("Setting default configuration...\n");

    /*
     * Default program configuration
     */
    set_log_level(PACO_LOG_INFO);
    set_log_stream(stdout);
    cfg.width = 16;
    cfg.stride = 4;
    cfg.lambda = 10; //nice number
    cfg.maxiter = 256;
    cfg.minchange = 1e-5;
    cfg.kappa = 0.95;
    cfg.weight_file = NULL;
    cfg.ref_file = NULL;
    cfg.init_file = NULL;
    cfg.iter_prefix = NULL;
    paco_info("Parsing arguments...\n");
    /*
     * call parser
     */
    argp_parse (&argp, argc, argv, 0, 0, &cfg);
    /*
     * print configuration for the record
     */
    const char* refstr = cfg.ref_file ? cfg.ref_file: "(none)";
    paco_info("About to run PACO-DCT:\n\tinput :%s\n\tmask  :%s\n\toutput:%s\n\tref.  :%s\n\twidth :%d\n\tstride:%d\n",
              cfg.input_file, cfg.mask_file, cfg.output_file, refstr, cfg.width, cfg.stride);
    if(cfg.stride>cfg.width) {
        paco_error("Stride must not exceed patch width");
        exit(1);
    }
    /*
     * load data.
     */
    paco_image_st* input = NULL;
    paco_image_st* init  = NULL;
    paco_info("Loading input data...\n");
    input = read_png_file(cfg.input_file);
    if(!input) {
        paco_error("Unable to read input!\n");
        exit(1);
    }
    /*
     * load mask file (mandatory) 
     */
    paco_image_st* mask = read_png_file(cfg.mask_file);
    if(!mask) {
        paco_error("Unable to read mask!\n");
        exit(1);
    }
    /*
     * load initial signal estimation, if available 
     */
    if(cfg.init_file) {
        paco_info("Loading initialisation...\n");
        init = read_png_file(cfg.init_file);
    }
    /*
     * load reference (aka ground truth) for quality assessment 
     */
    paco_image_st* ref;
    if (cfg.ref_file) {
        paco_info("Loading reference data ... \n");
        ref = read_png_file(cfg.ref_file);
    } else {
        ref = NULL;
    }
    paco_info("Input is %d x %d (x %d), bpp=%d\n",get_nrows(input), get_ncols(input), get_nchannels(input), get_bitd(input));

    const index_t N = get_ncols(input);
    const index_t C = get_nchannels(input);
    const index_t M = get_nrows(mask);
    /*
     * basic checks
     */
    if (get_nrows(mask) != get_nrows(input)) {
        paco_error("Error: number of rows in input and mask do not match.\n");
        exit(1);
    }
    if (get_ncols(mask) != get_ncols(input)) {
        paco_error("Error: number of columns in input and mask do not match.\n");
        exit(1);        
    }
    if (init && (get_ncols(init) != get_ncols(input))) {
        paco_error("Error: number of columns in initialization and input do not match.\n");
        exit(1);        
    }
    if (ref && (get_ncols(ref) != get_ncols(input))) {
        paco_error("Error: number of columns in reference and input do not match.\n");
        exit(1);        
    }
    /**
    * avoid any ground truth contamination from the start: set missing pixels to 0
    */
    for (index_t i = 0; i < M; i++) {
        for (index_t j = 0; j < N; j++) {
            if (get_sample(mask,i,j,0)) {
                for (index_t c = 0; c < C ; c++) {
                    set_sample(input,i,j,c,0);
                }
            }
        }
    }
    /*
    * preprocessing: pad signals if necessary, convert to YUV
    */
    index_t ph = compute_padded_size(M, cfg.width, cfg.stride);
    index_t pw = compute_padded_size(N, cfg.width, cfg.stride);

    paco_image_st* padded_input = pad_image_mirror(input, pw, ph);
    if (padded_input != input) {
        paco_image_free(input);
        free(input); input = 0;
    }
    paco_image_st* padded_mask = pad_image_mirror(mask, pw, ph); 
    if (padded_mask != mask) {
        paco_image_free(mask);
        free(mask); mask = 0;
    }
    paco_image_st* padded_ref = NULL;
    if(ref) {
        padded_ref = pad_image_mirror(ref, pw, ph);
        if (padded_ref != ref) {
            paco_image_free(ref);
            free(ref); ref = 0;
        }
    }
    paco_image_st* padded_init = NULL;
    if(init) {
        padded_init = pad_image_mirror(init, pw, ph);
        if (padded_init != init) {
            paco_image_free(init);
            free(init); init = 0;
        }
    }

    /*
     * perform colorspace conversion for color inputs */
    if (strcmp(get_colorspace(padded_input),COLORSPACE_RGB) == 0) {
        paco_image_convert_colorspace(padded_input,COLORSPACE_YUV); 
        if (padded_init) paco_image_convert_colorspace(padded_init,COLORSPACE_YUV);
        if (padded_ref) paco_image_convert_colorspace(padded_ref,COLORSPACE_YUV);
    } 
    paco_image_st* padded_output = paco_image_alloc(ph,pw,get_colorspace(padded_input));
    /*
     * instantiate and initialize the patch mapping information.
     * This contains the precomputed indexes of  all the important pixels and patches
     */
    paco_grid_mapping_init(get_channel_samples(padded_input,0), get_channel_samples(padded_mask,0), cfg.width, cfg.stride);
    paco_iterate_st* iter = NULL;
    /*
     * algorithm parameters taken  from the command line
     * see their description above
     */
    admm_params_st params;
    params.lambda0 = cfg.lambda;
    params.kappa = cfg.kappa;
    params.max_iter = cfg.maxiter;
    params.min_arg_change = cfg.minchange;
    params.min_cost_change = cfg.minchange;

    /*
     * run algorithm on each channel
     */
    const index_t nchannels = get_nchannels(padded_input);
    for (index_t c = 0; c < nchannels; c++) {
        /*
         * channels are identified by a single letter
         */
        const char channel_name = get_channel_name(padded_input,c);
        /*
         * get c-th channel from image
         */
        gsl_matrix_float* mask_samples = get_channel_samples(padded_mask,0);
        gsl_matrix_float* input_samples = get_channel_samples(padded_input,c);
        gsl_matrix_float* ref_samples = padded_ref != NULL? get_channel_samples(padded_ref,c) : NULL;
        gsl_matrix_float* output_samples = get_channel_samples(padded_output,c);
        gsl_matrix_float* init_samples = padded_init != NULL ? get_channel_samples(padded_init,c) : NULL;
        /*
         * if available, load precomputed DCT weights for this channel
         */
        gsl_vector_float* weights = NULL;
        if(cfg.weight_file) {
            char aux[128];
            snprintf(aux,128,"%s-%c.txt",cfg.weight_file,channel_name);
            paco_info("Loading DCT coefficients for channel %c from %s...\n",channel_name,aux);
            weights = read_weights(aux, cfg.width);
        }
        /*
         * create an instance of the PACO-DCT inpainting problem
         */
	if (cfg.iter_prefix) {
            char channel_iter_prefix[128];
            snprintf(channel_iter_prefix,128,"%s_channel_%c_",cfg.iter_prefix,channel_name);
            paco_problem_set_iter_prefix(channel_iter_prefix);
	}
	//
        if (c == 0) { 
            // first channel: perform full initialization
            paco_dct_inpainting_problem_init(mask_samples);
            // create iteration data
            iter = paco_iterate_alloc();
        }
        /*
         * set channel-specific info
         */
        paco_problem_set_input(input_samples);
        paco_problem_set_ref(ref_samples);
        paco_problem_set_weights(weights);
        gsl_vector_float_free(weights);
        /*
        * initialize iterate
        */
        paco_iter_init(iter,init_samples);
        /*
        * set ADMM algorithm parameters.
        */
        time(&t0);
        paco_info("lambda0=%f\n",params.lambda0);
        //
        // run ADMM
        //
        paco_admm(iter, &params);
        /*
        * save output.
        */
        time(&t1);
        /*
        * print stats for this channel
        */
        paco_info("\vChannel %c.\n", channel_name);
        paco_info("\ttime: %7.3ds.\n",t1-t0);
        if (padded_ref != NULL) {
            float* ref_data = ref_samples->data;
            float* out_data = iter->Y->data;
            float* mask_data = mask_samples->data;
            const index_t tda = input_samples->tda;
            const index_t ncols = input_samples->size2;
            const index_t nrows = input_samples->size1;
            double rmse = rmse_partial(ref_data, out_data, mask_data, nrows*ncols);
            double ssim =  iqa_ms_ssim(ref_data, out_data, mask_data, ncols, nrows, tda, NULL);
            paco_info("CHANNEL %d RMSE: %f      SSIM: %f \n", c, 255*rmse,  ssim);
        }
        gsl_matrix_float_memcpy(output_samples,iter->Y);
    }
    /*
     * free  problem module storage
     */
    paco_problem_destroy();
    /*
     * free mapping module storage
     */
    paco_mapping_destroy();
    /*
     * write result
     */
    if (strcmp(get_colorspace(padded_output),COLORSPACE_YUV) == 0) {
        paco_image_convert_colorspace(padded_output,COLORSPACE_RGB);
    }
    paco_image_clip(padded_output);
    paco_image_st* output = unpad_image(padded_output,N,M);
    write_png_file(cfg.output_file, output);
    if (output != padded_output) {
        paco_image_free(output);
        free(output); output = 0;
    }
    /*
     * free local storage.
     */
    paco_iterate_free(iter);
    paco_image_free(padded_output);
    free(padded_output);
    paco_image_free(padded_input);
    free(padded_input);
    paco_image_free(padded_mask);
    free(padded_mask);
    if (padded_ref) {
        paco_image_free(padded_ref);
	    free(padded_ref);
    }
    exit (0);
}


/*
 * argp callback for parsing a single option.
 */
static error_t parse_opt (int key, char *arg, struct argp_state *state) {
    /* Get the input argument from argp_parse,
     * which we know is a pointer to our arguments structure.
     */
    paco_config_st *cfg = (paco_config_st*) state->input;
    switch (key) {
    case 'q':
        set_log_level(PACO_LOG_ERROR);
        break;
    case 'v':
        set_log_level(PACO_LOG_DEBUG);
        break;
    case 'w':
        cfg->width = atoi(arg);
        break;
    case 'l':
        cfg->lambda = atof(arg);
        break;
    case 's':
        cfg->stride = atoi(arg);
        break;
    case 'm':
        cfg->maxiter = atoi(arg);
        break;
    case 'e':
        cfg->minchange = atof(arg);
        break;
    case 'k':
        cfg->kappa = atof(arg);
        break;
    case 'c':
        cfg->weight_file = arg;
        break;
    case 'p':
        cfg->iter_prefix = arg;
        break;
    case 'i':
        cfg->init_file = arg;
        break;

    case ARGP_KEY_ARG:
        switch (state->arg_num) {
        case 0:
            cfg->input_file = arg;
            break;
        case 1:
            cfg->mask_file = arg;
            break;
        case 2:
            cfg->output_file = arg;
            break;
        case 3:
            cfg->ref_file = arg;
            break;
        default:
            /** too many arguments! */
            paco_error("Too many arguments!.\n");
            argp_usage (state);
            break;
        }
        break;
    case ARGP_KEY_END:
        if (state->arg_num < 3) {
            /* Not enough mandatory arguments! */
            paco_error("Too FEW arguments!\n");
            argp_usage (state);
        }
        break;

    default:
        return ARGP_ERR_UNKNOWN;
    }
    return 0;
}
