/*
  * Copyright (c) 2019 Ignacio Francisco Ramírez Paulino and Ignacio Hounie
  * This program is free software: you can redistribute it and/or modify it
  * under the terms of the GNU Affero General Public License as published by
  * the Free Software Foundation, either version 3 of the License, or (at
  * your option) any later version.
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero
  * General Public License for more details.
  * You should have received a copy of the GNU Affero General Public License
  *  along with this program. If not, see <http://www.gnu.org/licenses/>.
*/
/**
 * \file paco_linalg.c
 * \brief implementation of some useful linear algebra functions 
 */
#include "paco_types.h"
#include "paco_linalg.h"
#include <omp.h>
#include <assert.h>

//
//============================================================================
//
double gsl_vector_float_sum ( const gsl_vector_float* v ) {
    double a = 0.0;
    const sample_t* pv = v->data;
    const size_t vs = v->stride;
    const size_t n = v->size;
#ifdef _OPENMP
    #pragma omp parallel for reduction(+:a)
#endif
    for ( size_t i = 0; i < n; ++i ) {
        a += *pv;
        pv += vs;
    }
    return a;
}

//
//============================================================================
//
double gsl_vector_float_dist2_squared ( const gsl_vector_float* u, const gsl_vector_float* v ) {
    assert ( u->size == v->size );
    register const size_t n = u->size;
    const sample_t* pu = u->data;
    const sample_t* pv = v->data;
    const size_t us = u->stride;
    const size_t vs = v->stride;
    double a = 0.0;
#ifdef _OPENMP
    #pragma omp parallel for reduction(+:a)
#endif
    for ( size_t i = 0; i < n ; ++i ) {
        const double d = pu[i*us]-pv[i*vs];
        a += d*d;
    }
    return a;
}

//
//============================================================================
//
double gsl_vector_float_dif_sum ( const gsl_vector_float* u, const gsl_vector_float* v ) {
    assert ( u->size == v->size );
    register const size_t n = u->size;
    const sample_t* pu = u->data;
    const sample_t* pv = v->data;
    const size_t us = u->stride;
    const size_t vs = v->stride;
    double d = 0.0;
#ifdef _OPENMP
    #pragma omp parallel for reduction(+:d)
#endif
    for (size_t i = 0; i < n; ++i) {
        d += *pu-*pv;
        pu += us;
        pv += vs;
    }
    return d;
}


//
//============================================================================
//
double gsl_matrix_float_sum ( const gsl_matrix_float* A ) {
    register const size_t m = A->size1;
    register const size_t n = A->size2;
    double a = 0.0;
    const sample_t* prow = A->data;
    for ( register size_t i = m ; i ; --i, prow += A->tda ) {
        const sample_t* pelem = prow;
#ifdef _OPENMP
        #pragma omp parallel for reduction(+:a)
#endif
        for ( size_t i = 0; i < n; ++i ) {
            a += pelem[i];
        }
    }
    return a;
}

//
//============================================================================
//
double gsl_matrix_float_dif_sum ( const gsl_matrix_float* A, const gsl_matrix_float* B ) {
    assert ( A->size1 == B->size1 );
    assert ( A->size2 == B->size2 );
    register const size_t m = A->size1;
    register const size_t n = A->size2;
    double a = 0.0;
    const sample_t* parow = A->data;
    const sample_t* pbrow = B->data;
    for ( register size_t i = m ; i ; --i, parow += A->tda, pbrow += B->tda ) {
        const sample_t* paelem = parow;
        const sample_t* pbelem = pbrow;
#ifdef _OPENMP
        #pragma omp parallel for reduction(+:a)
#endif
        for ( size_t i = 0; i < n; ++i ) {
            a += paelem[i] - pbelem[i];
        }
    }
    return a;
}

//
//============================================================================
//
/// Euclidean distance between two matrices as elements of R^{mxn}
/// @param A first matrix
/// @param B second matrix
/// @return the euclidean distance between A and B
double gsl_matrix_float_dist2_squared ( const gsl_matrix_float* A, const gsl_matrix_float* B ) {
    assert ( A->size1 == B->size1 );
    assert ( A->size2 == B->size2 );
    register const size_t m = A->size1;
    register const size_t n = A->size2;
    double a = 0.0;
    const sample_t* parow = A->data;
    const sample_t* pbrow = B->data;
    for ( register size_t i = m ; i ; --i, parow += A->tda, pbrow += B->tda ) {
        const sample_t* paelem = parow;
        const sample_t* pbelem = pbrow;
        double aj = 0.0;
#ifdef _OPENMP
        #pragma omp parallel for reduction(+:aj)
#endif
        for ( size_t i = 0; i < n; ++i ) {
            const double d = paelem[i] - pbelem[i];
            aj += d*d;
        }
        a += aj;
    }
    return a;
}


//
//============================================================================
//
double gsl_matrix_float_frobenius_squared(const gsl_matrix_float* A) {
    const size_t n = A->size1;
    double a = 0.0;
#ifdef _OPENMP
    #pragma omp parallel for reduction(+:a)
#endif
    for (size_t j = 0; j < n; ++j) {
        gsl_vector_float_const_view Aj = gsl_matrix_float_const_row(A,j);
        const double nAj = gsl_blas_snrm2(&Aj.vector);
        a += nAj*nAj;
    }
    return a;
}

//
//============================================================================
//
double gsl_matrix_float_frobenius(const gsl_matrix_float* A) {
    return sqrt(gsl_matrix_float_frobenius_squared(A));
}

//
//============================================================================
//
